import java.io.*;
import java.net.*;
import java.nio.*;
import java.nio.channels.*;
import java.util.*;

public class NonblockingIOSelectorServer {
  private static Map<SocketChannel, Queue<ByteBuffer>>
    pendingWrites = new HashMap<>();
  private static ByteBuffer readBuffer =
      ByteBuffer.allocateDirect(80);

  public static void main(String... args)
      throws IOException {
    ServerSocketChannel ssc = ServerSocketChannel.open();
    ssc.bind(new InetSocketAddress("localhost", 8080));
    ssc.configureBlocking(false);
    Selector selector = Selector.open();
    ssc.register(selector, SelectionKey.OP_ACCEPT);

    while(true) {
      selector.select(); // blocks until something happens
      Set<SelectionKey> keys = selector.selectedKeys();
      for (Iterator<SelectionKey> it = keys.iterator();
           it.hasNext(); ) {
        SelectionKey key = it.next();
        it.remove();
        if (key.isValid()) {
          if (key.isAcceptable()) {
            accept(key);
          } else if (key.isReadable()) {
            read(key);
          } else if (key.isWritable()) {
            write(key);
          }
        }
      }
    }
   }

  private static void write(SelectionKey key) throws IOException {
    SocketChannel socket = (SocketChannel) key.channel();
    System.out.println("Ready for writing: " + socket);

    // read buffers from queue.  If buffer is fully writen,
    // remove from queue, otherwise, change back to read
    Queue<ByteBuffer> queue = pendingWrites.get(socket);
    ByteBuffer writeBuffer;
    while((writeBuffer = queue.peek()) != null) {
      socket.write(writeBuffer);
      if (writeBuffer.hasRemaining()) {
        return;
      }
      queue.remove();
    }
    socket.register(key.selector(),
        SelectionKey.OP_READ);
  }

  private static void read(SelectionKey key)
      throws IOException {
    SocketChannel socket = (SocketChannel) key.channel();
    System.out.println("Reading from " + socket);
    readBuffer.clear();
    try {
      int bytesRead = socket.read(readBuffer);
      if (bytesRead == -1) {
        close(socket);
        return;
      }
      readBuffer.flip();
      ByteBuffer writeBuffer = ByteBuffer.allocate(
          readBuffer.limit()
      );
      while(readBuffer.hasRemaining()) {
        writeBuffer.put(
            (byte) Util.hokuspokus(readBuffer.get()));
      }
      writeBuffer.flip();
      pendingWrites.get(socket).add(writeBuffer);
      socket.register(key.selector(),
          SelectionKey.OP_WRITE);
    } catch (IOException e) {
      System.err.println(e);
      close(socket);
    }
  }

  private static void close(SocketChannel socket)
      throws IOException {
    socket.close();
    pendingWrites.remove(socket);
    System.out.println("Closed " + socket);
  }

  private static void accept(SelectionKey key)
      throws IOException {
    ServerSocketChannel ssc =
        (ServerSocketChannel) key.channel();
    SocketChannel newSocket = ssc.accept(); // nie null
    System.out.println("Connected to " + newSocket);
    newSocket.configureBlocking(false);
    pendingWrites.put(newSocket, new ArrayDeque<>());
    newSocket.register(key.selector(),
        SelectionKey.OP_READ);
  }
}
